import pkg_resources
import argparse
import mido
import sys


def version_info():
    versions = []
    for pkg in ("swood", "pillow-simd", "pillow", "PIL", "pyFFTW", "mido"):
        try:
            versions.append(str(pkg_resources.get_distribution(pkg)))
        except pkg_resources.DistributionNotFound:
            pass
    return " ".join(versions)


def swoodlive_installed():
    import importlib.util
    return importlib.util.find_spec("swoodlive") is not None


def run_cmd():
    parser = argparse.ArgumentParser(prog=("swood.exe" if sys.platform.startswith("win") else "swood"), description="swood.exe: the automatic ytpmv generator")

    parser.add_argument("in_wav", type=argparse.FileType("rb"), help="a short wav file to sample as the instrument")
    parser.add_argument("in_midi", type=mido.MidiFile, help="the MIDI to play with the wav sample")
    parser.add_argument("out_wav", type=argparse.FileType("wb"), help="path for the output wav file")

    parser.add_argument("--transpose", "-t", type=int, default=0, help="amount to transpose (semitones)")
    parser.add_argument("--speed", "-s", type=float, default=1.0, help="speed multiplier for the MIDI")
    parser.add_argument("--cachesize", "-c", type=float, default=7.5, help="wait time for cache notes the higher the quicker the render but the more memory it uses")
    parser.add_argument("--binsize", "-b", type=int, default=8192, help="FFT bin size; lower numbers make it faster but more off-pitch")

    parser.add_argument("--fullclip", "-f", action="store_true", help="always use the full sample without cropping")
    parser.add_argument("--optout", "-o", action="store_true", help="opt out of automatic bug reporting (or set the env variable SWOOD_OPTOUT)")

    if swoodlive_installed():
        parser.add_argument("--live", help="listen on a midi input and generate the output in realtime")

    parser.add_argument("--version", "-v", action="version", version=version_info(), help="get the versions of swood and its dependencies")

    args = parser.parse_args()

    from . import complain
    from . import midiparse
    from . import renderer
    from . import sample

    with complain.ComplaintFormatter():
        sample = sample.Sample(args.in_wav, args.binsize)
        midi = midiparse.MIDIParser(args.in_midi, sample, args.transpose, args.speed)
        renderer = renderer.NoteRenderer(sample, args.fullclip, args.cachesize)
        renderer.render(midi, args.out_wav)



if __name__ == "__main__":
    run_cmd()



from enum import Enum
import math

from PIL import Image
import progressbar
import numpy as np
from numpy import zeros, asarray, argmin, resize

from . import midiparse
from . import wavout

CachedNote = midiparse.CachedNote


class FileSaveType(Enum):
    ARRAY_TO_DISK = 0
    ARRAY_IN_MEM = 1
    SMART_CACHING = 2


class NoteRenderer:
    def __init__(self, sample, fullclip=False, cachesize=7.5, threshold=0.075):
        if threshold < 0:
            return ValueError("The threshold must be a positive number.")
        self.sample = sample
        self.fullclip = fullclip

        self.cachesize = cachesize * sample.framerate
        self.threshold = int(threshold * sample.framerate)

        self.notecache = {}

    def zoom(self, img, multiplier):
        return asarray(img.resize((int(round(img.size[0] * multiplier)), self.sample.channels), resample=Image.BICUBIC), dtype=np.int32)

    def render_note(self, note):
        scaled = self.zoom(self.sample.img, self.sample.maxfreq / note.pitch)
        if self.fullclip or len(scaled) < note.length + self.threshold:
            return scaled
        channels = self.sample.channels.shape[0]
        if channels == 1:
            cutoff = argmin(v + (d * 20) for d, v in enumerate(scaled[0][note.length:note.length + self.threshold]))
            return np.resize(scaled, (1, note.length + cutoff))
        else:
            cutoffs = [None] * channels
            for chan in range(channels):
                # find the nearest/closest zero crossing within the threshold and continue until that
                sample_end = np.empty(self.threshold)
                for distance, val in enumerate(scaled[chan][note.length:note.length + self.threshold]):
                    sample_end[distance] = (val) + (distance * 20)
                cutoffs[chan] = argmin(sample_end)
            merged_channels = zeros((channels, note.length + max(cutoffs)), dtype=np.int32)
            for chan in range(channels):
                cutoff = note.length + cutoffs[chan]
                merged_channels[chan][:cutoff] = scaled[chan][:cutoff]
            return merged_channels

    def render(self, midi, filename, pbar=True, savetype=FileSaveType.ARRAY_TO_DISK, clear_cache=True):
        if self.fullclip:
            # leave a small buffer at the end with space for one more sample
            end_buffer = int(math.ceil(midi.maxpitch * len(self.sample)))
        else:
            end_buffer = self.threshold  # it has to cut off sounds at the threshold anyway
        output_length = midi.length + end_buffer

        if savetype == FileSaveType.SMART_CACHING:
            #output = CachedWavFile(output_length, filename, self.sample.framerate)
            raise complain.ComplainToUser("Smart caching will be implemented in the future.")
        else:
            output = wavout.UncachedWavFile(output_length, filename, self.sample.framerate, self.sample.channels)

        if pbar:
            bar = progressbar.ProgressBar(widgets=[progressbar.Percentage(), " ", progressbar.Bar(), " ", progressbar.ETA()], max_value=midi.notecount)
            progress = 0

        caching = self.cachesize > 0

        if caching:
            tick = 8

        for time, notes in midi.notes:
            for note in notes:
                if hash(note) in self.notecache:
                    rendered_note = self.notecache[hash(note)]
                    rendered_note.used += 1  # increment the used counter each time for the "GC" below
                else:
                    rendered_note = CachedNote(time, self.render_note(note))
                    self.notecache[hash(note)] = rendered_note
                note_volume = (note.volume / midi.maxvolume) * self.sample.volume
                output.add_data(time, (rendered_note.data * note_volume).astype(np.int32))

                if pbar:
                    # increment progress bar
                    progress += 1
                    bar.update(progress)

            
            if caching:
                # cache "garbage collection":
                # if a CachedNote is more than 7.5 (default) seconds old it removes it from the cache to save mem(e)ory
                tick += 1
                if tick == 15:
                    tick = 0
                    for k in list(self.notecache.keys()):
                        if time - self.notecache[k].length > self.cachesize and self.notecache[k].used < 3:
                            del self.notecache[k]

        if clear_cache:
            self.notecache.clear()

        if savetype == FileSaveType.ARRAY_IN_MEM:
            return output.channels
        else:
            output.save()



import traceback
import sys
import os


class ComplainToUser(Exception):
    pass


class ComplaintFormatter(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc, tb):
        if exc_type == ComplainToUser:
            print("Error: {}".format(exc))
        elif isinstance(exc, Exception):
            tb = traceback.format_exc()
            if "--optout" in sys.argv or "-o" in sys.argv or os.environ.get("SWOOD_OPTOUT") is not None:
                print("Something went wrong. A bug report will not be sent because of your environment variable/CLI option.")
                traceback.print_exc()
            else:
                print("Something went wrong. A bug report will be sent to help figure it out. (see --optout)")
                try:
                    import http.client
                    conn = http.client.HTTPSConnection("meme.institute")
                    conn.request("POST", "/swood/bugs/submit", tb)
                    resp = conn.getresponse().read().decode("utf-8")
                    if resp == "done":
                        print("New bug submitted!")
                    elif resp == "dupe":
                        print("This bug is already in the queue to be fixed.")
                    else:
                        raise Exception
                except Exception:
                    print("Apparently we can't even send a bug report right. Sorry about that.")
                    traceback.print_exc()
        return True



import numpy as np
from . import complain
import wave


class CachedWavFile:
    pass

    
class UncachedWavFile:
    def __init__(self, length, filename, framerate, channels=1, dtype=np.int32):
        self.channels = np.zeros((channels, length), dtype=dtype)
        self.framerate = framerate
        self.filename = filename

    def add_data(self, start, data):
        # cut it off at the end of the file in case of cache shenanigans
        length = min(self.channels.shape[1] - start, data.shape[1])
        for chan in range(self.channels.shape[0]):
            self.channels[chan][start:start + length] += data[chan][:length]

    def save(self):
        # we can't use a with statement with wave.open here
        # because wave.open throws a really weird error I can't catch
        try:
            with (open(self.filename, "wb") if isinstance(self.filename, str) else self.filename) as wavfile:
                with wave.Wave_write(self) as wav:
                    wav.initfp(wavfile)
                    wav.setparams((self.channels.shape[0],  # channels
                                   self.channels.dtype.itemsize,  # sample width
                                   self.framerate,  # sample rate
                                   self.channels.shape[1],  # number of frames
                                   "NONE", "not compressed"))  # compression type
                    wav.writeframesraw(self.channels.reshape(self.channels.size, order="F"))
        except IOError:
            raise complain.ComplainToUser("Can't save output file '{}'.".format(self.filename))



from PIL import Image
import numpy as np
from . import complain
import pyfftw
import wave

pyfftw.interfaces.cache.enable()


class CalculatedFFT:
    def __init__(self, avgdata, spacing):
        self.avgdata = avgdata
        self.spacing = spacing


class Sample:
    def __init__(self, filename, binsize, volume=0.8, delete_raw_data=True):
        self.binsize = binsize

        if binsize < 2:
            raise complain.ComplainToUser("FFT bin size must be at least 2.")

        self.delete_raw = delete_raw_data  # delete raw data after FFT analysis
        self._maxfreq = None
        self._fft = None

        self.wav = self.parse_wav(filename)

        max_amplitude = max(max(abs(min(chan)), abs(max(chan))) for chan in self.wav)
        self.volume = 256 ** 4 / (max_amplitude * 2) * volume
        self.img = Image.frombytes("I",
                                   (self.length, self.channels),
                                   self.wav.astype(np.int32).tobytes(),
                                   "raw", "I", 0, 1)
        # Pillow recommends those last args because of a bug in the raw parser
        # See http://pillow.readthedocs.io/en/3.2.x/reference/Image.html?highlight=%22raw%22#PIL.Image.frombuffer

    def parse_wav(self, filename):
        try:
            with wave.open(filename, "rb") as wavfile:
                self.sampwidth = wavfile.getsampwidth()
                self.framerate = wavfile.getframerate()
                self.channels = wavfile.getnchannels()
                self.length = wavfile.getnframes()

                if self.sampwidth == 1:
                    self.size = np.int8
                elif self.sampwidth == 2:
                    self.size = np.int16
                elif self.sampwidth == 3 or self.sampwidth == 4:
                    self.size = np.int32
                else:
                    raise wave.Error

                wav = np.zeros((self.channels, self.length), dtype=self.size)
                for i in range(0, self.length):
                    frame = wavfile.readframes(1)
                    for chan in range(self.channels):
                        wav[chan][i] = int.from_bytes(frame[self.sampwidth * chan:self.sampwidth * (chan + 1)], byteorder="little", signed=True)
                return wav
        except IOError:
            raise complain.ComplainToUser("Error opening WAV file at path '{}'.".format(filename))
        except wave.Error:
            raise complain.ComplainToUser("This WAV type is not supported. Try opening the file in Audacity and exporting it as a standard WAV.")

    @property
    def fft(self):
        if not self._fft:
            if self.binsize % 2 != 0:
                print("Warning: Bin size must be a multiple of 2, correcting automatically")
                self.binsize += 1
            spacing = float(self.framerate) / self.binsize
            avgdata = np.zeros(self.binsize // 2, dtype=np.float64)
            for chan in range(self.channels):
                for i in range(0, self.wav.shape[1], self.binsize):
                    data = np.array(self.wav[chan][i:i + self.binsize], dtype=self.size)
                    if len(data) != self.binsize:
                        continue
                    fft = pyfftw.interfaces.numpy_fft.fft(data)
                    fft = np.abs(fft[:self.binsize // 2])
                    avgdata += fft
                    del data
                    del fft
            if max(avgdata) == 0:
                print("Warning: Bin size is too large to analyze sample; dividing by 2 and trying again")
                self.binsize = self.binsize // 2
                self._fft = self.fft
            else:
                if self.delete_raw:
                    del self.wav
                self._fft = CalculatedFFT(avgdata, spacing)
        return self._fft

    @property
    def maxfreq(self):
        if not self._maxfreq:
            self._maxfreq = (np.argmax(self.fft.avgdata[1:]) * self.fft.spacing) + (self.fft.spacing / 2)
        return self._maxfreq

    def __len__(self):
        return self.length



from __future__ import print_function
import sys

if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 4):
    print("Sorry, swood.exe requires at least Python 3.4 to run correctly.")
    sys.exit(1)

from setuptools import setup
import importlib.util
import importlib
import platform

import os
import io

try:
    import pip
except ImportError:
    print("Installing pip...")
    import ensurepip
    ensurepip.bootstrap()
    importlib.invalidate_caches()
    import pip


def install_package(pkg):
    print("Installing {}...".format(pkg))
    tmp_stdout = sys.stdout
    sys.stdout = io.StringIO()
    try:
        pip.main(["install", pkg])
    except SystemExit as e:
        stdout_str = sys.stdout.getvalue()
        sys.stdout = tmp_stdout
        if e.code != 0:
            print("pip failed to install {}.\nYou may need a working C compiler to continue.\npip stdout:".format(pkg))
            print(stdout_str)
            raise ImportError
        else:
            importlib.invalidate_caches()


def get_flags():
    sse4 = False
    avx2 = False
    try:
        # something feels very wrong about running assembly in python
        # most of this taken from https://github.com/workhorsy/py-cpuinfo

        sixtyfour = sys.maxsize > 2 ** 32  # the docs say platform.machine() can be unreliable

        def asm_func(restype=None, argtypes=(), byte_code=[]):
            byte_code = bytes.join(b'', byte_code)
            address = None
            # Allocate a memory segment the size of the byte code
            size = len(byte_code)
            address = ctypes.pythonapi.valloc(size)
            if not address:
                raise Exception("Failed to valloc")
            # Mark the memory segment as writeable only
            WRITE = 0x2
            if ctypes.pythonapi.mprotect(address, size, WRITE) < 0:
                raise Exception("Failed to mprotect")
            # Copy the byte code into the memory segment
            if ctypes.pythonapi.memmove(address, byte_code, size) < 0:
                raise Exception("Failed to memmove")
            # Mark the memory segment as writeable and executable only
            WRITE_EXECUTE = 0x2 | 0x4
            if ctypes.pythonapi.mprotect(address, size, WRITE_EXECUTE) < 0:
                raise Exception("Failed to mprotect")
            # Cast the memory segment into a function
            functype = ctypes.CFUNCTYPE(restype, *argtypes)
            fun = functype(address)
            return fun, address

        def run_asm(*byte_code):
            # Convert the byte code into a function that returns an int
            func, address = asm_func(ctypes.c_uint64 if sixtyfour else ctypes.c_uint32, (), byte_code)
            # Call the byte code like a function
            retval = func()
            size = ctypes.c_size_t(len(byte_code))
            # Free the function memory segment
            # Remove the executable tag on the memory
            READ_WRITE = 0x1 | 0x2
            if ctypes.pythonapi.mprotect(address, size, READ_WRITE) < 0:
                raise Exception("Failed to mprotect")
            ctypes.pythonapi.free(address)
            return retval

        def is_bit_set(reg, bit):
            mask = 1 << bit
            return reg & mask > 0

        def eax():
            if sixtyfour:
                return b"\x66\xB8\x01\x00"  # mov eax,0x1"
            else:
                return (b"\x31\xC0"         # xor ax,ax
                        b"\x40")            # inc ax

        ecx = run_asm(eax(),
                      b"\x0f\xa2"         # cpuid
                      b"\x89\xC8"         # mov ax,cx
                      b"\xC3")            # ret

        ebx = run_asm(
                b"\xB8\x01\x00\x00\x80"  # mov ax,0x80000001
                b"\x0f\xa2"         # cpuid
                b"\x89\xD8"         # mov ax,bx
                b"\xC3")            # ret

        sse4 = is_bit_set(ecx, 19) or is_bit_set(ecx, 20)

        avx2 = is_bit_set(ebx, 5)  # because a midi synthesizer definitely needs to support server-grade hardware
    except Exception:
        pass
    return (sse4, avx2)

install = False
simd = False

if len(sys.argv) > 1 and "install" in sys.argv:
    pkgs = [package.project_name.lower() for package in pip.get_installed_distributions()]

    if "pillow-simd" in pkgs:
        print("pillow-simd is already installed. swood will install with SIMD support.")
        install = False
        simd = True
    if platform.machine() in ("i386", "x86_64") and os.name() == "posix":
        try:
            import ctypes
            sse4, avx2 = get_flags()
            if avx2:
                print("Your processor supports AVX2. swood will install with SIMD support.")
                os.environ["CFLAGS"] = "-mavx2"
                install = True
                simd = True
            elif sse4:
                print("Your processor supports SSE4. swood will install with SIMD support.")
                install = True
                simd = True
        except:
            simd = False

    # we need to install numpy first because pyfftw needs it and pip has bad dependency resolution
    # and we need to set this flag because pyFFTW uses a deprecated API apparently
    os.environ["CFLAGS"] = "-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION"

    for pkg in ("numpy", "pyfftw", "progressbar2"):
        if pkg not in pkgs:
            install_package(pkg)

    if install:
        try:
            install_package("pillow-simd")
        except ImportError:
            print("SIMD support failed to install. swood will run slower.")
            simd = False

reqs = ['mido', 'numpy', 'progressbar2', 'pyfftw']

if not simd:
    reqs.append('pillow')

setup(
    name='swood',
    version='0.9.9',
    description='With just one sample and a MIDI you too can make YTPMVs',
    long_description='Are you tired of manually pitch-adjusting every sound for your shitposts? Toil no more with auto-placement of sound samples according to a MIDI!',
    url='https://meme.institute/swood',
    author='Milkey Mouse',
    author_email='milkeymouse@meme.institute',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='swood memes youtubepoop ytp ytpmvs',
    packages=["swood"],

    install_requires=reqs,

    entry_points={
        'console_scripts': [
            'swood=swood:run_cmd',
        ],
    },
)



import collections
import operator

import mido

from . import complain


class Note:
    def __init__(self, length=None, pitch=None, volume=None, starttime=None):
        self.starttime = starttime
        self.length = length
        self.volume = volume
        self.pitch = pitch

    def __hash__(self):
        return hash((self.length, self.pitch))

    def __eq__(self, other):
        return self.length == other.length and self.pitch == other.pitch

    def __len__(self):
        return len(self.data)


class CachedNote:
    def __init__(self, length, rendered):
        self.used = 1
        self.length = length
        self.data = rendered

    def __len__(self):
        return len(self.data)


class MIDIParser:
    def __init__(self, filename, wav, transpose=0, speed=1):  # TODO: convert the rest of the function to the new notes
        if speed <= 0:
            return ValueError("The speed must be a positive number.")
        results = collections.defaultdict(list)
        notes = collections.defaultdict(list)
        self.notecount = 0
        self.maxvolume = 0
        self.maxpitch = 0
        volume = 0

        try:
            with (mido.MidiFile(filename, "r") if isinstance(filename, str) else filename) as mid:
                time = 0
                for message in mid:
                    time += message.time
                    if "channel" in vars(message) and message.channel == 10:
                        continue  # channel 10 is reserved for percussion
                    if message.type == "note_on":
                        note = Note()
                        note.starttime = int(round(time * wav.framerate / speed))
                        note.volume = 127 if message.velocity == 0 else message.velocity
                        note.pitch = self.note_to_freq(message.note + transpose)

                        notes[message.note].append(note)
                        volume += note.volume
                        self.maxvolume = max(volume, self.maxvolume)
                    elif message.type == "note_off":
                        note = notes[message.note].pop(0)
                        if len(notes[message.note]) == 0:
                            del notes[message.note]

                        try:
                            results[note.starttime].append(note)
                        except IndexError:
                            print("Warning: There was a note end event at {} seconds with no matching begin event".format(time))

                        self.notecount += 1
                        volume -= note.volume
                        self.maxpitch = max(self.maxpitch, note.pitch)
                        note.length = int(time * wav.framerate / speed) - note.starttime

                if len(notes) != 0:
                    print("Warning: The MIDI ended with notes still playing, assuming they end when the MIDI does")
                    for ntime, nlist in notes.items():
                        for note in nlist:
                            note.length = int(time * wav.framerate / speed) - note.starttime
                            self.notecount += 1

                if self.notecount == 0:
                    raise complain.ComplainToUser("This MIDI file doesn't have any notes in it!")

                self.notes = sorted(results.items(), key=operator.itemgetter(0))
                self.length = max(max(note.starttime + note.length for note in nlist) for _, nlist in self.notes)
        except IOError:
            raise complain.ComplainToUser("Error opening MIDI file '{}'.".format(filename))
        except IndexError:
            raise complain.ComplainToUser("This MIDI file is broken. Try opening it in MidiEditor (https://meme.institute/midieditor) and saving it back out again.")

    def note_to_freq(self, notenum):
        # see https://en.wikipedia.org/wiki/MIDI_Tuning_Standard
        return (2.0 ** ((notenum - 69) / 12.0)) * 440.0
